import { Chessboard, isEmpty, Square, squareAtPosition, emptyfile,createInitialChessboard } from "./chessboard";
import { Move } from "./movements";
import { equals, left, right, top, bottom } from "./position";

/**
 * Check that the file squares are empty or not 
 * @param chessboard 
 * @param move 
 */
function isEmptyFile(chessboard : Chessboard, move : Move) : boolean {
    let start: number;
    let end: number;
    let file : number = move.from!.file;

    if (file !== move.to!.file) { 
        // Should not happen
        return false;
    }

    if (move.from!.rank > move.to!.rank) { //if the starting square > the destination square
        start = move.to!.rank;
        end = move.from!.rank;
    } else {
        end = move.to!.rank;
        start = move.from!.rank;       
    }
    

    let i : number = start+1; // We only want to check the "middle" squares so we don't take in account the strating square
    while (i < end && chessboard.board[file][i].isEmpty) 
    {        
        i++;
    }

    return i === end; 
}

/**
 * Check the rank squares are empty or not
 * @param chessboard
 * @param move
 */
function isEmptyRank(chessboard : Chessboard, move: Move): boolean {
    let start:  number;
    let end : number;
    let rank : number = move.from!.rank;

    if (rank !== move.to!.rank) { 
        // Should not happen
        return false;
    }

    if (move.from!.file > move.to!.file) { //if the square of origin > square of destination so
        start = move.to!.file;
        end = move.from!.file;
    } else {
        end = move.to!.file;
        start = move.from!.file;       
    }

    let i : number = start+1; // We only want to check the "middle" squares so we don't take in account the strating square
    while (i < end && chessboard.board[i][rank].isEmpty) 
    { 
          i++;
    }

    return i === end; 
}
/**
 * Check if squares in diagonal are empty or not
 * @param chessboard
 * @param move
 */
function isEmptyDiag(chessboard: Chessboard, move: Move): boolean{
    let startRank: number = move.from!.rank; 
    let startFile: number = move.from!.file;
    let endRank: number = move.to!.rank;
    let endFile: number = move.to!.file;

    // Diagonal at top right
    if (startFile < endFile && startRank > endRank)
    {
        startFile++
        startRank--
        while (startFile < endFile && startRank > endRank && chessboard.board[startFile][startRank].isEmpty)
        {
            startFile++
            startRank--
        }
    }
    // Diagonal at bottom left
    if (startFile> endFile && startRank < endRank)
    {
        startFile--
        startRank++
        while (startFile > endFile && startRank < endRank && chessboard.board[startFile][startRank].isEmpty)
        {
            startFile--
            startRank++
        }
    }
    // Diagonal at top left
    if (startFile < endFile && startRank < endRank)
    {
        startFile++
        startRank++
        while ( startFile < endFile && startRank < endRank && chessboard.board[startFile][startRank].isEmpty){
            startFile++
            startRank++
        }
    }
    // Diagonal at bottom right
    if (startFile > endFile && startRank > endRank){
        startFile--
        startRank--
        while (startFile > endFile && startRank > endRank && chessboard.board[startFile][startRank].isEmpty){
            startFile--
            startRank--
        }
    }
    return (startFile == endFile && startRank == endRank);
}



/**
 * Checks whether a Black Pawn can perform a given move.
 * A pawn can move forward to the unoccupied square immediately in front of 
 * it on the same file, or on its first move it can advance two squares along 
 * the same file, provided both squares are unoccupied (black dots in the 
 * diagram); or the pawn can capture an opponent's piece on a square diagonally 
 * in front of it on an adjacent file, by moving to that square (black "x"s). 
 * 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function blackPawnMove(board: Chessboard, move: Move): boolean {

    if (equals(move.to!, bottom(move.from!))) {
        //console.log("Single forward");
        return isEmpty(board, move.to!);
    }

    if (move.from!.rank === 6 && equals(move.to!, bottom(bottom(move.from!)))) {
        //console.log("Double forward");
        return isEmpty(board, bottom(move.from!)) && isEmpty(board, move.to!);
    }

    if (equals(move.to!, left(bottom(move.from!))) || equals(move.to!, right(bottom(move.from!)))) {
        let destination: Square = squareAtPosition(board, move.to!);
        return !(destination.isEmpty || !destination.piece!.isWhite)
    }

    return false;
}

/**
 * A pawn can move forward to the unoccupied square immediately in front of 
 * it on the same file, or on its first move it can advance two squares along 
 * the same file, provided both squares are unoccupied (black dots in 
 * the diagram); or the pawn can capture an opponent's piece on a square diagonally 
 * in front of it on an adjacent file, by moving to that square (black "x"s). 
 * 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function whitePawnMove(board: Chessboard, move: Move): boolean {

    if (equals(move.to!, top(move.from!))) {
        return isEmpty(board, move.to!);
    }

    if (move.from!.rank === 1 && equals(move.to!, top(top(move.from!)))) {
        return isEmpty(board, top(move.from!)) && isEmpty(board, move.to!);
    }

    if (equals(move.to!, left(top(move.from!))) || equals(move.to!, right(top(move.from!)))) {
        let destination: Square = squareAtPosition(board, move.to!);
        return !(destination.isEmpty || destination.piece!.isWhite)
    }

    return false;
}

/**
 * Checks whether a King can perform a given move.
 * The king moves one square in any direction. 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function kingMove(board: Chessboard, move: Move): boolean {
   
    // Check of movement to move one square in any direction
    if (equals(move.to!, bottom(move.from!)) || equals(move.to!, top(move.from!))
     || equals(move.to!, right(move.from!)) || equals(move.to!, left(move.from!))
     || equals(move.to!, top(right(move.from!))) || equals(move.to!, top(left(move.from!)))
     || equals(move.to!, bottom(left(move.from!))) || equals(move.to!, bottom(right(move.from!))))
    {
         let destination : Square = squareAtPosition(board,move.to!);
         let origin : Square = squareAtPosition(board, move.from!);
         
         // Check that "destination square is empty or thath the color of the piece that is on the origin square
         // is different than the color of the piece which is on the destination square" is true
         return (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);
    }
    return false;
}

/**
 * Checks whether a Queen can perform a given move.
 * The queen combines the power of a rook and bishop and can move any 
 * number of squares along a rank, file, or diagonal, but cannot leap over other pieces.
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function queenMove(board: Chessboard, move: Move): boolean {
    let destination : Square = squareAtPosition(board, move.to!);
    let origin : Square = squareAtPosition(board, move.from!);

    // Validation of movement to move horizontally 
    // Check that there is no pawn between origin square and the destination square
    if  (move.from!.rank === move.to!.rank && isEmptyRank(board,move)==true){

        return (destination.isEmpty || destination.piece!.isWhite !== origin.piece!.isWhite);
    }

    // Validation of movement to move vertically
    // Check that there is no pawn between origin square and the destination square
    if  (move.from!.file ===move.to!.file && isEmptyFile(board,move)==true) {

        return (destination.isEmpty || destination.piece!.isWhite !== origin.piece!.isWhite);   
    }

    // Validation of movement to move diagonally
    const filedifference : number = origin.position.file - destination.position.file;
    const rankdifference : number = origin.position.rank - destination.position.rank;

    // Check that the absolute value of differences is equal, which lets us know that the movement
    // is a diagonal movement and that there is no pawn between the origin square and the destination square
    if ((Math.abs(filedifference) == Math.abs(rankdifference)) && (isEmptyDiag(board, move)==true)){

        return (destination.isEmpty || destination.piece!.isWhite !== origin.piece!.isWhite);
    }

    return false;
}

/**
 * Checks whether a Empress can perform a given move.
 * An Empress can move any number of squares along a rank or file, 
 * but cannot leap over other pieces.
 * An Empress can also move to any of the closest squares that are not on the 
 * same rank, file, or diagonal. (Thus the move forms an "L"-shape: 
 * two squares vertically and one square horizontally, or two 
 * squares horizontally and one square vertically.) 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function empressMove(board: Chessboard, move: Move): boolean {
   let destination : Square = squareAtPosition(board, move.to!);
   let origin : Square = squareAtPosition(board, move.from!);

   // Validation of movement to move vertically
   if( move.from!.file ===move.to!.file && isEmptyFile(board,move)==true)
   {   
         return  (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);     
   }

   // Validation of movement to move horizontally 
   if (move.from!.rank === move.to!.rank && isEmptyRank(board,move)==true)
   { 
         return (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);      
   } 
   
   //Validation of movement to move in "L"
   if(equals(move.to!, left(top(top(move.from!)))) 
     || equals(move.to!, left(bottom(bottom(move.from!)))) 
     || equals(move.to!, right(top(top(move.from!)))) || equals(move.to!, right(bottom(bottom(move.from!)))) 
     || equals(move.to!, top(left(left(move.from!)))) || equals(move.to!, top(right(right(move.from!)))) ||
     equals(move.to!, bottom(left(left(move.from!)))) || equals(move.to!, bottom(right(right(move.from!))))  
    )
   {     
         return  (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);    
   }

return false;
}

/**
 * Checks whether a Princess can perform a given move.
 * A princess can move any number of squares diagonally, 
 * but cannot leap over other pieces.
 * A princess can also move to any of the closest squares that are not on the 
 * same rank, file, or diagonal. (Thus the move forms an "L"-shape: 
 * two squares vertically and one square horizontally, or two 
 * squares horizontally and one square vertically.) 
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function princessMove(board: Chessboard, move: Move): boolean {
    let destination: Square = squareAtPosition(board, move.to!);
    let origin: Square = squareAtPosition(board, move.from!);
    const fileDifference : number = origin.position.file - destination.position.file;
    const rankDifference : number = origin.position.rank - destination.position.rank;
    
    // Validation of movement to move diagonally
    if ((Math.abs(fileDifference) == Math.abs(rankDifference)) && (isEmptyDiag(board, move)==true))
    {
         return (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);
    } 

    // Validation of movement to move in "L"
    if (equals(move.to!, left(top(top(move.from!)))) ||
     equals(move.to!, left(bottom(bottom(move.from!)))) ||
     equals(move.to!, right(top(top(move.from!)))) ||
     equals(move.to!, right(bottom(bottom(move.from!)))) ||
     equals(move.to!, top(left(left(move.from!)))) ||
     equals(move.to!, top(right(right(move.from!)))) ||
     equals(move.to!, bottom(left(left(move.from!)))) ||      
     equals(move.to!, bottom(right(right(move.from!))))      
     )
    {
        return (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);
    }  

    return false;
}

/**
 * Checks whether a Camel can perform a given move.
 * The Camel move forms an "L"-shape: 
 * three squares vertically and one square horizontally, or three 
 * squares horizontally and one square vertically.) 
 * 
 * The camel can leap over other pieces.
 * 
 * @param board The chessboard of the current game
 * @param move 
 */
export function camelMove(board: Chessboard, move: Move): boolean {
    let destination : Square = squareAtPosition(board,move.to!);
    let origin : Square = squareAtPosition(board, move.from!);
    
    // Validation of movement to move in "L"
    if (equals(move.to!, left(top(top(top(move.from!))))) || equals(move.to!, left(bottom(bottom(bottom(move.from!)))))
     || equals(move.to!, right(top(top(top(move.from!))))) || equals(move.to!, right(bottom(bottom(bottom(move.from!)))))
     || equals(move.to!, top(right(right(right(move.from!))))) || equals(move.to!, top(left(left(left(move.from!)))))
     || equals(move.to!, bottom(right(right(right(move.from!))))) || equals(move.to!, bottom(left(left(left(move.from!)))))
     )
    {    
         return (destination.isEmpty||destination.piece!.isWhite !== origin.piece!.isWhite);
    }
    return false;
}   