import { Expect, Test, Setup} from "alsatian";
import * as isPossible from '../../main/ts/move-validation'
import * as pieces from '../../main/ts/piece'
import { Chessboard, createEmptyChessboard, putPiece, pieceAtPosition } from '../../main/ts/chessboard';
import {Position, position} from '../../main/ts/position';
import { Move, move } from '../../main/ts/movements';

let chessboard : Chessboard;

const positionA4 : Position = position(0, 3) // A4
const positionA5 : Position = position(0, 4) // A5
const positionA6 : Position = position(0, 5) // A6
const positionA7 : Position = position(0, 6) // A7
const positionA8 : Position = position(0, 7) // A8

const positionB1 : Position = position(1, 0) // B1
const positionB2 : Position = position(1, 1) // B2
const positionB3 : Position = position(1, 2) // B3
const positionB4 : Position = position(1, 3) // B4
const positionB5 : Position = position(1, 4) // B5
const positionB6 : Position = position(1, 5) // B6

const positionC3 : Position = position(2, 2) // C3
const positionC4 : Position = position(2, 3) // C4
const positionC5 : Position = position(2, 4) // C5
const positionC6 : Position = position(2, 5) // C6
const positionC7 : Position = position(2, 6) // C7

const positionD1 : Position = position(3, 0) // D1
const positionD2 : Position = position(3, 1) // D2
const positionD3 : Position = position(3, 2) // D3
const positionD4 : Position = position(3, 3) // D4
const positionD5 : Position = position(3, 4) // D5
const positionD6 : Position = position(3, 5) // D6
const positionD7 : Position = position(3, 6) // D7


const positionE1 : Position = position(4, 0) // E1
const positionE3 : Position = position(4, 2) // E3
const positionE4 : Position = position(4, 3) // E4
const positionE6 : Position = position(4, 5) // E6
const positionE8 : Position = position(4, 7) // E8

const positionF1 : Position = position(5, 0) // F1
const positionF2 : Position = position(5, 1) // F2
const positionF6 : Position = position(5, 5) // F6
const positionF7 : Position = position(5, 6) // F7

const positionG3 : Position = position(6, 2) // G3
const positionG4 : Position = position(6, 3) // G4
const positionG5 : Position = position(6, 4) // G5
const positionG6 : Position = position(6, 5) // G6

const positionH1 : Position = position(7, 0) // H1
const positionH3 : Position = position(7, 2) // H3
const positionH4 : Position = position(7, 3) // H4
const positionH5 : Position = position(7, 4) // H5
const positionH7 : Position = position(7, 6) // H7

// Tree square horizontally and one square vertically 
const moveE4_B3 : Move = move(positionE4, positionB3);
const moveE4_B5 : Move = move(positionE4, positionB5);
const moveE4_H3 : Move = move(positionE4, positionH3);
const moveE4_H5 : Move = move(positionE4, positionH5);

// Tree square vertically and one square horizontally 
const moveE4_F1 : Move = move(positionE4, positionF1);
const moveE4_F7 : Move = move(positionE4, positionF7);
const moveE4_D1 : Move = move(positionE4, positionD1);
const moveE4_D7 : Move = move(positionE4, positionD7);

// Impossible moves
const moveE4_B4 : Move = move(positionE4, positionB4);
const moveE4_D4 : Move = move(positionE4, positionD4);
const moveE4_D5 : Move = move(positionE4, positionD5);
const moveE4_E1 : Move = move(positionE4, positionE1);
const moveE4_E3 : Move = move(positionE4, positionE3);
const moveE4_E6 : Move = move(positionE4, positionE6);
const moveE4_G4 : Move = move(positionE4, positionG4);
const moveE4_G6 : Move = move(positionE4, positionG6);
const moveE4_H1 : Move = move(positionE4, positionH1);

export class TestCamelMoves {
    @Setup
    beforeEach() {
        // Initialize an empty chessboard
        chessboard = createEmptyChessboard ();
        // Place a white Camel on E4
        putPiece(chessboard, positionE4, pieces.whiteCamel)
    }

    @Test("A Camel can move three squares horizontally and one square vertically")
    testCanMoveThreeHorizontalAndOneVertical() {
        // Check the following moves are possible: moveE4_H3, moveE4_H5, moveE4_B3, moveE4_B5
        Expect(isPossible.camelMove(chessboard, moveE4_B3)).toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_B5)).toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_H3)).toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_H5)).toBeTruthy();
    }

    @Test("A Camel can move three squares vertically and one square horizontally")
    testCanMoveThreeVerticalAndOneHorizontal() {
        // Check the following moves are possible: moveE4_F1, moveE4_F7, moveE4_D1, moveE4_D7
        Expect(isPossible.camelMove(chessboard, moveE4_F1)).toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_F7)).toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_D1)).toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_D7)).toBeTruthy();
    }

    @Test("A Camel can leap other pieces")
    testCanLeapOtherPieces() {
        // Place a black Pawn on E1
        // Place a white Pawn on D2
        putPiece(chessboard, positionE1, pieces.blackPawn);
        putPiece(chessboard, positionD2, pieces.whitePawn);
        let mvtCanLeap: Move = {from: positionE4, to: positionD1, isValid: true}
        // Check the move moveE4_D1 is possible
        Expect(isPossible.camelMove(chessboard, mvtCanLeap)).toBeTruthy();
    }

    @Test("A Camel cannot move diagonally")
    testCannotMoveDiagonally() {
        // Check the following moves are impossible: moveE4_D5, moveE4_H1, moevE4_G6
        Expect(isPossible.camelMove(chessboard, moveE4_D5)).not.toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_G6)).not.toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_H1)).not.toBeTruthy();
    }

    @Test("A Camel cannot move horizontally")
    testCannotMoveHorizontally() {
        // Check the following moves are impossible : moveE4_D4, moveE4_B4, moveE4_G4
        Expect(isPossible.camelMove(chessboard, moveE4_D4)).not.toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_B4)).not.toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_G4)).not.toBeTruthy();
    }

    @Test("A Camel cannot move vertically")
    testCannotMoveVertically() {
        // Check the following moves are impossible : moveE4_E1, moveE4_E3, moveE4_E6
        Expect(isPossible.camelMove(chessboard, moveE4_E1)).not.toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_E3)).not.toBeTruthy();
        Expect(isPossible.camelMove(chessboard, moveE4_E6)).not.toBeTruthy();

    }

    @Test("A Camel can capture a piece from another color")
    testCanCaptureAnotherColor() {
        // Place a black Pawn on D1
        putPiece(chessboard, positionD1, pieces.blackPawn);
        let mvtCapture: Move = {from: positionE4, to: positionD1, isValid: true}
        // Check the move moveE4_D1 is possible
        Expect(isPossible.camelMove(chessboard, mvtCapture)).toBeTruthy();
    }

    @Test("A Camel cannot capture a piece from the same color")
    testCannotCaptureSameColor() {
        // Place a white Pawn on F1
        putPiece(chessboard, positionF1, pieces.whitePawn);
        let mvtCapture: Move = {from: positionE4, to: positionF1, isValid: true}
        // Check the move moveE4_F1 is impossible
        Expect(isPossible.camelMove(chessboard, mvtCapture)).not.toBeTruthy();
    }
}


